import sys
import json
import argparse

from stitcher.stitcher import Stitcher
from stitcher.api import deploy

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("call_graph",
        nargs="*",
        help="Paths to call graphs to be stitched together in JSON format")
    parser.add_argument(
        "--simple",
        action="store_true",
        help="Output in simple format",
        default=False
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output path",
        default=None
    )
    parser.add_argument(
        "-a",
        "--api",
        action="store_true",
        help="Deploy the server",
        default=None
    )

    args = parser.parse_args()

    if args.api:
        deploy()
        return

    stitcher = Stitcher(args.call_graph, args.simple)
    stitcher.stitch()

    output = json.dumps(stitcher.output())
    if args.output:
        with open(args.output, "w+") as f:
            f.write(output)
    else:
        print (output)

if __name__ == "__main__":
    main()



#
# Copyright (c) 2018-2021 FASTEN.
#
# This file is part of FASTEN
# (see https://www.fasten-project.eu/).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from stitcher.node import Node

class CallGraph:
    def __init__(self, cg):
        self.cg = cg
        self.internal_calls = []
        self.external_calls = []
        self.nodes = {}
        self.id_to_node = {}

        self.product = self.cg["product"]
        self.version = self.cg["version"]

        self._parse_cg()

    def get_node(self, modname, name):
        if not self.nodes.get(modname, None):
            return None
        return self.nodes[modname].get(name, None)

    def get_internal_calls(self):
        return self.internal_calls

    def get_external_calls(self):
        return self.external_calls

    def _parse_cg(self):
        def iterate_mods(d, internal):
            for mod_id, data in d.items():
                if internal:
                    mod_node = Node(mod_id, product=self.product,
                                    version=self.version)
                    modname = mod_node.get_modname()
                else:
                    modname = mod_id
                self.nodes[modname] = {}

                for id, info in data["namespaces"].items():
                    super_cls = None
                    if info["metadata"].get("superClasses", None) != None:
                        super_cls = []
                        for cls in info["metadata"]["superClasses"]:
                            super_cls.append(Node(cls, product=self.product,
                                                        version=self.version))

                    node = Node(info["namespace"], product=self.product,
                                super_cls=super_cls, version=self.version)

                    self.id_to_node[id] = node
                    if node.get_modname():
                        self.nodes[node.get_modname()][node.get_callable()] = node

        def iterate_calls(calls):
            res = []
            for src, dst, metadata in calls:
                if self.id_to_node.get(src, None) and self.id_to_node.get(dst, None):
                    res.append([
                        self.id_to_node.get(src),
                        self.id_to_node.get(dst)])
            return res

        iterate_mods(self.cg["modules"]["internal"], True)
        iterate_mods(self.cg["modules"]["external"], False)

        self.internal_calls = iterate_calls(self.cg["graph"]["internalCalls"])
        self.external_calls = iterate_calls(self.cg["graph"]["externalCalls"])



import json

from flask import Flask, jsonify, request
from stitcher.stitcher import Stitcher


app = Flask(__name__)


@app.route("/stitch", methods = ['POST'])
def stitch():
    content = request.json

    stitcher = Stitcher([], False)
    for cg in content.values():
        stitcher.parse_cg(cg)
    stitcher.stitch()

    output = json.dumps(stitcher.output())

    return output


def deploy(host='0.0.0.0', port=5000):
    app.run(threaded=True, host=host, port=port)



#
# Copyright (c) 2018-2021 FASTEN.
#
# This file is part of FASTEN
# (see https://www.fasten-project.eu/).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import sys
import json

from stitcher.cg import CallGraph

class Stitcher:
    def __init__(self, call_graph_paths, simple):
        self.simple = simple
        self.cgs = {}
        self.id_cnt = 0
        self.node_to_id = {}
        self.stitched = {
            "edges": [],
            "nodes": {}
        }

        self.nodes_cnt = 0
        self.edges_cnt = 0
        self.resolved_cnt = 0
        self.edges_cnt_no_builtin = 0

        self._parse_cgs(call_graph_paths)

    def stitch(self):
        for product, cg in self.cgs.items():
            internal_calls = cg.get_internal_calls()
            external_calls = cg.get_external_calls()
            self.edges_cnt += len(internal_calls) + len(external_calls)
            self.edges_cnt_no_builtin += len(internal_calls) + len(external_calls)
            self.resolved_cnt += len(internal_calls)

            for src, dst in internal_calls:
                self._assign_id(src.to_string(self.simple))
                self._assign_id(dst.to_string(self.simple))
                self.stitched["edges"].append([
                    self.node_to_id[src.to_string(self.simple)],
                    self.node_to_id[dst.to_string(self.simple)],
                ])

            for src, dst in external_calls:
                if ".builtin" in dst.to_string():
                    self.edges_cnt_no_builtin -= 1
                for resolved in self._resolve(dst):
                    self.resolved_cnt += 1
                    self._assign_id(src.to_string(self.simple))
                    self._assign_id(resolved.to_string(self.simple))
                    self.stitched["edges"].append([
                        self.node_to_id[src.to_string(self.simple)],
                        self.node_to_id[resolved.to_string(self.simple)],
                    ])

        self.nodes_cnt = self.id_cnt
        for node, id in self.node_to_id.items():
            self.stitched["nodes"][id] = {"URI": node, "metadata": {}}

    def output(self):
        return self.stitched

    def _parse_cgs(self, paths):
        for p in paths:
            with open(p, "r") as f:
                cg = json.load(f)
                if self.cgs.get(cg["product"], None):
                    continue

                self.cgs[cg["product"]] = CallGraph(cg)

    def parse_cg(self, cg_json):
        if not self.cgs.get(cg_json["product"], None):
            self.cgs[cg_json["product"]] = CallGraph(cg_json)

    def _resolve(self, node, search_parents=True):
        product = node.get_product()
        callbl = node.get_callable().split(".")

        # if we don't have the call graph for that product
        # we cannot resolve any calls
        if self.cgs.get(product.replace("_", "-")):
            product = product.replace("_", "-")

        if not self.cgs.get(product):
            return None

        for i in range(1, len(callbl)):
            modname = ".".join(callbl[:i])
            fnname = ".".join(callbl[i:])

            actual = self.cgs[product].get_node(modname, fnname)

            if not actual:
                if search_parents:
                    parent_fnname = ".".join(callbl[i:-1])
                    if self.cgs[product].get_node(modname, parent_fnname):
                        fn = self._resolve_mro(product, modname, parent_fnname, callbl[-1])
                        if fn:
                            yield fn
            elif actual.is_func or actual.is_class:
                yield actual

    def _resolve_mro(self, product, modname, cls, name):
        if not self.cgs.get(product, None):
            return None

        node = self.cgs[product].get_node(modname, cls)

        resolved = None
        for parent in node.get_class_hier():
            if parent.get_product() == product:
                resolved = self.cgs[product].get_node(
                    parent.get_modname(),
                    parent.get_callable() + "." + name)
            else:
                for parent_resolved in self._resolve(parent, search_parents=False):
                    if parent_resolved:
                        resolved = self.cgs[parent_resolved.get_product()].get_node(
                            parent_resolved.get_modname(),
                            parent_resolved.get_callable() + "." + name)

            if resolved:
                return resolved

        return None

    def _err_and_exit(self, msg):
        print (msg)
        sys.exit(1)

    def _assign_id(self, node_str):
        if not self.node_to_id.get(node_str, None):
            self.node_to_id[node_str] = self.id_cnt
            self.id_cnt += 1



from setuptools import setup, find_packages

def get_long_desc():
    with open("README.md", "r") as readme:
        desc = readme.read()

    return desc

def setup_package():
    setup(
        name='pycg-stitch',
        version='0.0.2',
        description='Stitcher for FASTEN Python call graphs',
        long_description=get_long_desc(),
        long_description_content_type="text/markdown",
         url='https://github.com/fasten-project/pycg-stitch',
        license='Apache Software License',
        packages=find_packages(),
        install_requires=['flask'],
        entry_points = {
            'console_scripts': [
                'pycg-stitch=stitcher.__main__:main',
            ],
        },
        classifiers=[
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3'
        ],
        author = 'Vitalis Salis, Giorgos Drosos',
        author_email = 'vitsalis@gmail.com, drosos007@gmail.com'
    )

if __name__ == '__main__':
    setup_package()



#
# Copyright (c) 2018-2021 FASTEN.
#
# This file is part of FASTEN
# (see https://www.fasten-project.eu/).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
class Node:
    def __init__(self, uri_str, product=None, super_cls=None, version=None):
        self.uri_str = uri_str
        self.product = product
        self.version = version
        self.super_cls = super_cls or []

        if len(uri_str.split("/")) == 5:
            self.internal = False
            self.external = True
        else:
            self.internal = True
            self.external = False

        self.modname = self.callable = ""

        splitted = self.uri_str.split("/")
        if self.internal:
            self.modname = splitted[1]
            self.callable = splitted[2]
        else:
            self.product = splitted[2]
            self.callable = splitted[4]

        self.is_class = super_cls != None
        self.is_func = False
        if self.callable.endswith("()"):
            self.is_func = True
            self.callable = self.callable[:-2]

    def get_product(self):
        return self.product

    def get_version(self):
        return self.version

    def get_modname(self):
        return self.modname

    def get_callable(self):
        return self.callable

    def get_class_hier(self):
        return self.super_cls

    def to_string(self, simple=False):
        uri = ""
        if not simple:
            uri += "fasten:"
        uri += "//" + self.product
        if self.version and not simple:
            uri += "$" + self.version
        uri += "/" + self.modname + "/" + self.callable

        if self.is_func:
            uri += "()"
        return uri
